]> git.ipfire.org Git - thirdparty/mdadm.git/blobdiff - monitor.c
Improve partition table code.
[thirdparty/mdadm.git] / monitor.c
index bd67e1ad1e29798a117181cacc4c5d51ad89164e..e43e545ce196f1370ad68b78c9b02dac67ae89f9 100644 (file)
--- a/monitor.c
+++ b/monitor.c
@@ -1,3 +1,22 @@
+/*
+ * mdmon - monitor external metadata arrays
+ *
+ * Copyright (C) 2007-2009 Neil Brown <neilb@suse.de>
+ * Copyright (C) 2007-2009 Intel Corporation
+ *
+ * This program is free software; you can redistribute it and/or modify it
+ * under the terms and conditions of the GNU General Public License,
+ * version 2, as published by the Free Software Foundation.
+ *
+ * This program is distributed in the hope it will be useful, but WITHOUT
+ * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
+ * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
+ * more details.
+ *
+ * You should have received a copy of the GNU General Public License along with
+ * this program; if not, write to the Free Software Foundation, Inc.,
+ * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
+ */
 
 #include "mdadm.h"
 #include "mdmon.h"
@@ -47,21 +66,20 @@ static int read_attr(char *buf, int len, int fd)
        return n;
 }
 
-int get_resync_start(struct active_array *a)
+static unsigned long long read_resync_start(int fd)
 {
        char buf[30];
        int n;
 
-       n = read_attr(buf, 30, a->resync_start_fd);
+       n = read_attr(buf, 30, fd);
        if (n <= 0)
-               return n;
-
-       a->resync_start = strtoull(buf, NULL, 10);
-
-       return 1;
+               return 0;
+       if (strncmp(buf, "none", 4) == 0)
+               return MaxSector;
+       else
+               return strtoull(buf, NULL, 10);
 }
 
-
 static enum array_state read_state(int fd)
 {
        char buf[20];
@@ -180,42 +198,48 @@ static int read_and_act(struct active_array *a)
        int check_degraded = 0;
        int deactivate = 0;
        struct mdinfo *mdi;
+       int dirty = 0;
 
        a->next_state = bad_word;
        a->next_action = bad_action;
 
        a->curr_state = read_state(a->info.state_fd);
        a->curr_action = read_action(a->action_fd);
+       a->info.resync_start = read_resync_start(a->resync_start_fd);
        for (mdi = a->info.devs; mdi ; mdi = mdi->next) {
                mdi->next_state = 0;
-               if (mdi->state_fd >= 0)
+               if (mdi->state_fd >= 0) {
+                       mdi->recovery_start = read_resync_start(mdi->recovery_fd);
                        mdi->curr_state = read_dev_state(mdi->state_fd);
+               }
        }
 
        if (a->curr_state <= inactive &&
            a->prev_state > inactive) {
                /* array has been stopped */
-               get_resync_start(a);
                a->container->ss->set_array_state(a, 1);
                a->next_state = clear;
                deactivate = 1;
        }
        if (a->curr_state == write_pending) {
-               get_resync_start(a);
                a->container->ss->set_array_state(a, 0);
                a->next_state = active;
+               dirty = 1;
        }
        if (a->curr_state == active_idle) {
                /* Set array to 'clean' FIRST, then mark clean
                 * in the metadata
                 */
                a->next_state = clean;
+               dirty = 1;
        }
        if (a->curr_state == clean) {
-               get_resync_start(a);
                a->container->ss->set_array_state(a, 1);
        }
-
+       if (a->curr_state == active ||
+           a->curr_state == suspended ||
+           a->curr_state == bad_word)
+               dirty = 1;
        if (a->curr_state == readonly) {
                /* Well, I'm ready to handle things.  If readonly
                 * wasn't requested, transition to read-auto.
@@ -226,11 +250,12 @@ static int read_and_act(struct active_array *a)
                        /* explicit request for readonly array.  Leave it alone */
                        ;
                } else {
-                       get_resync_start(a);
                        if (a->container->ss->set_array_state(a, 2))
                                a->next_state = read_auto; /* array is clean */
-                       else
+                       else {
                                a->next_state = active; /* Now active for recovery etc */
+                               dirty = 1;
+                       }
                }
        }
 
@@ -242,7 +267,6 @@ static int read_and_act(struct active_array *a)
                 * until the array goes inactive or readonly though.
                 * Just check if we need to fiddle spares.
                 */
-               get_resync_start(a);
                a->container->ss->set_array_state(a, a->curr_state <= clean);
                check_degraded = 1;
        }
@@ -342,7 +366,7 @@ static int read_and_act(struct active_array *a)
        if (deactivate)
                a->container = NULL;
 
-       return 1;
+       return dirty;
 }
 
 static struct mdinfo *
@@ -414,6 +438,7 @@ static int wait_and_act(struct supertype *container, int nowait)
        struct active_array *a, **ap;
        int rv;
        struct mdinfo *mdi;
+       static unsigned int dirty_arrays = ~0; /* start at some non-zero value */
 
        FD_ZERO(&rfds);
 
@@ -442,17 +467,25 @@ static int wait_and_act(struct supertype *container, int nowait)
                ap = &(*ap)->next;
        }
 
-       if (manager_ready && *aap == NULL) {
-               /* No interesting arrays. Lets see about exiting.
-                * Note that blocking at this point is not a problem
-                * as there are no active arrays, there is nothing that
-                * we need to be ready to do.
+       if (manager_ready && (*aap == NULL || (sigterm && !dirty_arrays))) {
+               /* No interesting arrays, or we have been told to
+                * terminate and everything is clean.  Lets see about
+                * exiting.  Note that blocking at this point is not a
+                * problem as there are no active arrays, there is
+                * nothing that we need to be ready to do.
                 */
-               int fd = open(container->device_name, O_RDONLY|O_EXCL);
+               int fd = open_dev_excl(container->devnum);
                if (fd >= 0 || errno != EBUSY) {
                        /* OK, we are safe to leave */
-                       dprintf("no arrays to monitor... exiting\n");
-                       remove_pidfile(container->devname);
+                       if (sigterm && !dirty_arrays)
+                               dprintf("caught sigterm, all clean... exiting\n");
+                       else
+                               dprintf("no arrays to monitor... exiting\n");
+                       if (!sigterm)
+                               /* On SIGTERM, someone (the take-over mdmon) will
+                                * clean up
+                                */
+                               remove_pidfile(container->devname);
                        exit_now = 1;
                        signal_manager();
                        exit(0);
@@ -464,7 +497,7 @@ static int wait_and_act(struct supertype *container, int nowait)
                sigprocmask(SIG_UNBLOCK, NULL, &set);
                sigdelset(&set, SIGUSR1);
                monitor_loop_cnt |= 1;
-               rv = pselect(maxfd+1, &rfds, NULL, NULL, NULL, &set);
+               rv = pselect(maxfd+1, NULL, NULL, &rfds, NULL, &set);
                monitor_loop_cnt += 1;
                if (rv == -1 && errno == EINTR)
                        rv = 0;
@@ -486,7 +519,11 @@ static int wait_and_act(struct supertype *container, int nowait)
                container->ss->sync_metadata(container);
        }
 
+       rv = 0;
+       dirty_arrays = 0;
        for (a = *aap; a ; a = a->next) {
+               int is_dirty;
+
                if (a->replaces && !discard_this) {
                        struct active_array **ap;
                        for (ap = &a->next; *ap && *ap != a->replaces;
@@ -499,8 +536,17 @@ static int wait_and_act(struct supertype *container, int nowait)
                        /* FIXME check if device->state_fd need to be cleared?*/
                        signal_manager();
                }
-               if (a->container)
-                       rv += read_and_act(a);
+               if (a->container) {
+                       is_dirty = read_and_act(a);
+                       rv |= 1;
+                       dirty_arrays += is_dirty;
+                       /* when terminating stop manipulating the array after it
+                        * is clean, but make sure read_and_act() is given a
+                        * chance to handle 'active_idle'
+                        */
+                       if (sigterm && !is_dirty)
+                               a->container = NULL; /* stop touching this array */
+               }
        }
 
        /* propagate failures across container members */